import datetime
import os
import pathlib
import posixpath
from typing import Optional

from django.contrib.postgres.fields import ArrayField
from django.core.exceptions import SuspiciousFileOperation
from django.core.files.storage import Storage, default_storage
from django.db.models import ImageField
from django.db.models.fields.files import FieldFile

from .forms import ImageListFormField


def validate_file_name(name, allow_relative_path=False):
    # TODO: when drop support for django 3.2, remove this function
    # and import from django

    # Remove potentially dangerous names
    if os.path.basename(name) in {"", ".", ".."}:
        raise SuspiciousFileOperation("Could not derive file name from '%s'" % name)

    if allow_relative_path:
        # Use PurePosixPath() because this branch is checked only in
        # FileField.generate_filename() where all file paths are expected to be
        # Unix style (with forward slashes).
        path = pathlib.PurePosixPath(name)
        if path.is_absolute() or ".." in path.parts:
            raise SuspiciousFileOperation(
                "Detected path traversal attempt in '%s'" % name
            )
    elif name != os.path.basename(name):
        raise SuspiciousFileOperation("File name '%s' includes path elements" % name)

    return name


class ImageListField(ArrayField):
    def __init__(
        self,
        *args,
        max_length: int = 150,
        storage: Optional[Storage] = None,
        upload_to: str = "",
        **kwargs,
    ):
        self.max_length = max_length or 150
        self.storage = storage or default_storage
        self.upload_to = upload_to or ""
        kwargs["base_field"] = ImageField(
            max_length=self.max_length, upload_to=upload_to
        )
        super().__init__(
            *args,
            **kwargs,
        )

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        path = "image_uploader_widget.postgres.fields.ImageListField"
        kwargs.update(
            {
                "base_field": self.base_field.clone(),
                "size": self.size,
                "max_length": self.max_length,
                "upload_to": self.upload_to,
            }
        )
        return name, path, args, kwargs

    def generate_filename(self, instance, filename):
        """
        Apply (if callable) or prepend (if a string) upload_to to the filename,
        then delegate further processing of the name to the storage backend.
        Until the storage layer, all file paths are expected to be Unix style
        (with forward slashes).
        """
        if callable(self.upload_to):
            filename = self.upload_to(instance, filename)
        else:
            dirname = datetime.datetime.now().strftime(str(self.upload_to))
            filename = posixpath.join(dirname, filename)
        filename = validate_file_name(filename, allow_relative_path=True)
        return self.storage.generate_filename(filename)

    def _get_file(self, instance, file):
        if isinstance(file, str):
            return FieldFile(instance, self, file)

        field_file = FieldFile(instance, self, file.name)
        field_file.file = file
        field_file._committed = False
        return field_file

    def pre_save(self, model_instance, add):
        value = super().pre_save(model_instance, add)
        value = [self._get_file(model_instance, item) for item in value]

        for file in value:
            if file and not file._committed:
                file.save(file.name, file.file, save=False)

        return value

    def save_form_data(self, instance, data):
        if not data:
            data = []
        return super().save_form_data(instance, data)

    def formfield(self, **kwargs):
        return super().formfield(
            **{
                "form_class": ImageListFormField,
                **kwargs,
            }
        )



from django import forms
from django.contrib import admin
from django.db import models


class ImageUploaderInlineWidget(forms.ClearableFileInput):
    template_name = (
        "image_uploader_widget/admin/inline_image_uploader_preview_widget.html"
    )


class ImageUploaderInline(admin.StackedInline):
    template = "image_uploader_widget/admin/inline_image_uploader.html"
    extra = 0
    add_image_text = ""
    drop_text = ""
    empty_text = ""
    empty_icon = ""
    drop_icon = ""
    add_icon = ""
    accept = "image/*"

    formfield_overrides = {models.ImageField: {"widget": ImageUploaderInlineWidget()}}

    def get_add_image_text(self):
        return self.add_image_text

    def get_drop_text(self):
        return self.drop_text

    def get_empty_text(self):
        return self.empty_text

    def get_empty_icon(self):
        return self.empty_icon

    def get_drop_icon(self):
        return self.drop_icon

    def get_add_icon(self):
        return self.add_icon

    def get_accept(self):
        return self.accept

    def get_formset(self, request, obj=None, **kwargs):
        item = super(ImageUploaderInline, self).get_formset(request, obj, **kwargs)
        item.add_image_text = self.get_add_image_text()
        item.drop_text = self.get_drop_text()
        item.empty_text = self.get_empty_text()
        item.empty_icon = self.get_empty_icon()
        item.drop_icon = self.get_drop_icon()
        item.add_icon = self.get_add_icon()
        item.accept = self.get_accept()
        return item

    @property
    def media(self):
        return forms.Media(
            js=[
                "image_uploader_widget/js/image-uploader-modal.js",
                "image_uploader_widget/js/image-uploader-inline.js",
            ],
            css={
                "screen": [
                    "image_uploader_widget/css/image-uploader-inline.css",
                ]
            },
        )


class OrderedImageUploaderInline(ImageUploaderInline):
    template = "image_uploader_widget/admin/ordered_inline_image_uploader.html"
    order_field = "order"

    def get_order_field(self, request):
        return self.order_field

    def get_formset(self, request, obj=None, **kwargs):
        item = super(OrderedImageUploaderInline, self).get_formset(
            request, obj, **kwargs
        )
        item.order_field = self.get_order_field(request)
        return item

    def get_queryset(self, request):
        queryset = super().get_queryset(request)
        field = self.get_order_field(request)
        return queryset.order_by(field)

    @property
    def media(self):
        return forms.Media(
            js=[
                "image_uploader_widget/js/vendor/sortable.min.js",
                "image_uploader_widget/js/image-uploader-modal.js",
                "image_uploader_widget/js/image-uploader-inline.js",
            ],
            css={
                "screen": [
                    "image_uploader_widget/css/image-uploader-inline.css",
                ]
            },
        )



import json
from typing import Any, List

from django import forms
from django.contrib.postgres.forms import SplitArrayWidget
from django.core.files.storage import default_storage
from django.core.files.uploadedfile import InMemoryUploadedFile


class ImageUploaderArrayWidget(SplitArrayWidget):
    template_name = "image_uploader_widget/postgres/image_array.html"
    drop_text = ""
    empty_text = ""
    empty_icon = ""
    drop_icon = ""

    def get_drop_text(self):
        return self.drop_text

    def get_empty_text(self):
        return self.empty_text

    def get_empty_icon(self):
        return self.empty_icon

    def get_drop_icon(self):
        return self.drop_icon

    def __init__(
        self, drop_text="", empty_text="", empty_icon="", drop_icon="", **kwargs
    ):
        self.drop_text = drop_text
        self.empty_text = empty_text
        self.empty_icon = empty_icon
        self.drop_icon = drop_icon

        widget = forms.ClearableFileInput()
        super().__init__(widget, 0, **kwargs)

    def _get_image(self, path):
        if isinstance(path, InMemoryUploadedFile):
            return None
        return default_storage.url(path)

    def get_files_from_value(self, value: Any) -> List[str]:
        return [self._get_image(name) for name in value]

    def get_context(self, name, value, attrs=None):
        value_raw = value or []
        value = list(filter(None, self.get_files_from_value([*value_raw])))
        self.size = len(value)

        context = super(ImageUploaderArrayWidget, self).get_context(name, value, attrs)
        if not context:
            context = {}

        for i in range(0, len(value)):
            context["widget"]["subwidgets"][i]["value"] = value[i]
            context["widget"]["subwidgets"][i]["value_raw"] = value_raw[i]

        return {
            **context,
            "inline_admin_formset": {
                "inline_formset_data": json.dumps(
                    {
                        "name": context["widget"]["name"],
                        "options": {
                            "prefix": context["widget"]["name"],
                        },
                    }
                ),
                "formset": {
                    "prefix": context["widget"]["name"],
                },
            },
            "custom": {
                "drop_text": self.get_drop_text(),
                "empty_text": self.get_empty_text(),
                "empty_icon": self.get_empty_icon(),
                "drop_icon": self.get_drop_icon(),
            },
        }

    def value_from_datadict(self, data, files, name):
        total_forms = int(data.get("images-TOTAL_FORMS"))
        result = []
        for i in range(0, total_forms):
            image_file = files.get(f"images-{i}-image")
            image_raw = data.get(f"images-{i}-RAW")
            image_delete = data.get(f"images-{i}-DELETE")
            if image_delete:
                continue

            if image_file:
                result = [*result, image_file]
            else:
                result = [*result, image_raw]

        return result

    @property
    def needs_multipart_form(self):
        return True

    @property
    def media(self):
        return forms.Media(
            js=(
                "image_uploader_widget/js/vendor/sortable.min.js",
                "image_uploader_widget/js/image-uploader-modal.js",
                "image_uploader_widget/js/image-uploader-inline.js",
            ),
            css={
                "screen": ("image_uploader_widget/css/image-uploader-inline.css",),
            },
        )



from itertools import chain

from django import forms
from django.contrib.postgres.utils import prefix_validation_error
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _

from .widget import ImageUploaderArrayWidget


class ImageListFormField(forms.Field):
    default_error_messages = {
        "item_invalid": _("Item %(name)s in the array did not validate:"),
    }

    def __init__(self, **kwargs):
        kwargs.pop("base_field")
        self.max_length = kwargs.pop("max_length") or 150

        self.required = False
        self.base_field = forms.ImageField(max_length=self.max_length)
        widget = ImageUploaderArrayWidget()
        kwargs.setdefault("widget", widget)
        super().__init__(**kwargs)

    def to_python(self, value):
        value = super().to_python(value)
        return [self.base_field.to_python(item) for item in value]

    def clean(self, value):
        cleaned_data = []
        errors = []

        max_size = len(value)
        for index in range(max_size):
            item = value[index]
            if isinstance(item, str):
                cleaned_data.append(item)
                continue

            file_name = item.name
            try:
                cleaned_item = self.base_field.clean(item)
                cleaned_data.append(cleaned_item)
            except ValidationError as error:
                errors.append(
                    prefix_validation_error(
                        error,
                        self.error_messages["item_invalid"],
                        code="item_invalid",
                        params={"name": file_name},
                    )
                )
                cleaned_data.append(None)
            else:
                errors.append(None)

        errors = list(filter(None, errors))
        if errors:
            raise ValidationError(list(chain.from_iterable(errors)))
        return cleaned_data

    def has_changed(self, initial, data):
        try:
            data = self.to_python(data)
        except ValidationError:
            pass
        else:
            if initial in self.empty_values and data in self.empty_values:
                return False
        return super().has_changed(initial, data)
